"Python3 program using common n-grams to identify formulaic sequences."

##  FORMULEX :
##
##  Python3 program to discover formulaic sequences :
##  Copyright 2015, R.S. Forsyth.
##  first version : 09/02/2015
##  last revision : 30/05/2016

ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math, os, string, sys
import  random, time


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  paracopy, parashow
from glib import  showobj
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  add2dic
from quaylib import  bastats, ranx
from quaylib import  phi4, philite
##from quaylib import  corsums, corrval
from quaylib import  fivenum, midvalue
from quaylib import  addsnips
from quaylib import  getkeys, keyshow
from quaylib import  getfrex
from quaylib import  outpars
from quaylib import  subtuple


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","formulex"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfanum initial
    ("mindocf",3,0,10),  ##  minimum worthwhile document frequency (??used)
    ("snipsize",115,4,2048),  ##  number of tokens in a snippet (57=ssss)
    ("maxiglen",5,2,9),  ##  maximum # of tokens in an n-gram
    ("miniglen",2,1,4),  ##  minimum n-gram size 2use in extending n-grams
    ("topgrams",100,8,4096),  ##  number of most frequent n-grams to keep
    ("dropsubs",0,0,1),  ##  whether to drop shorter subgrams on o/p (??used)
    ("outforms",100,10,256),  ##  number of top forms to write out
    ("metafile",''),  ##  training corpus
    ("trainmet",''),  ##  ditto (takes precedence)
    ("testmeta",''),  ##  test metafile:
    ("blocfile",''),  ##  machine-readable results file
    ("formfile",''),  ##  human-readable results file
    ("gramdata",''),  ##  file for input of n-grams (from outgrams.py)
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"formulex.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "1.5"
Gold = 5.0**0.5*0.5+0.5
Mindocs = 4
Gramtail = "_gram"
Formtail = "_food"  ##  formulaic ordering of documents
Bloctail = "_flab"  ##  frequent lexically assembled bundles
Longname = 28  ##  longest class name; longer will be truncated
Maxtoks = 16 * 1024  ##  will only seek tokens in front chunk of text file


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if paradat.miniglen > paradat.maxiglen:
        paradat.miniglen = max(1,paradat.maxiglen//2)
        print("miniglen adjusted to",paradat.miniglen)
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of main (training) corpus :
    if paradat.trainmet:  ##  preferred case, trainmet takes precedence
        paradat.metafile = paradat.trainmet
        print(paradat.metafile,"to be used to indicate training texts.")
    elif paradat.metafile:
        print(paradat.metafile,"to be used to indicate training texts.")
    ##  else error, caught just below.
    metaname = getattr(paradat,"metafile",'')
    if not metaname:
        print("Can't proceed without metafile!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
        if paradat.cats < 2:
            print("Must have at least 2 categories !!")
            print("Program halts.")
            return  -2
    else:
        print("No. of text types =",paradat.cats)

    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    paradat.midsize = midvalue(paradat.sizelist)
    paradat.minsize = min(paradat.sizelist)
    print("Shortest =",paradat.minsize,"tokens.")
    print("median size =",paradat.midsize)
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("mean size =",round(paradat.meansize,2))
    print("Longest =",paradat.maxsize,"tokens.")
    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -5
    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].totchars = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].tottoks)
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  reference class is largest
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)
    print("reference category : ",paradat.refcode)

    for c in paradat.catlist:
        print(c,catinfo[c].docs)
        ##  showobj(catinfo[c])
        ##  catinfo doesn't have freqtabs @present.

    ##  attach extra attributes to doclist :
    allfrex = {} ; allsnips = {}
    paradat.totsnips = addsnips(paradat,doclist)
    for d in doclist:
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  attach category number as well as class name.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)  ##  each text has freqtab attached
        d.vocsize = len(d.freqtab)
        ##  also make overall freqtabs :
        add2dic(allfrex,d.freqtab)
        for snip in d.snipsets:  ##  also tabulate snippet frequency
            for v in snip:
                sf = allsnips.get(v,0)
                allsnips[v] = sf + 1
        ##  need to add into catinfo ??
        d.text = ' '.join(d.toklist) + ' '
        d.textlen = len(d.text)  ##  trailing blank included
        catinfo[d.outcome].totchars += d.textlen  ##  total character count
    print("snippets =",paradat.totsnips)
    paradat.doclist = doclist  ##  copy reference
    paradat.allfrex = allfrex  ##  ditto
    paradat.vocsize = len(allfrex)

    ##  prepare o/p files :
    if paradat.blocfile:
        blocname = paradat.blocfile
    else:
        blocname = paradat.outpath + paradat.jobname + Bloctail + ".dat"
    paradat.dumpfile = open(blocname,'w',encoding="utf8")
    ##  also need separate results file :
    if paradat.formfile:
        listfile = paradat.formfile
    else:
        listfile = paradat.outpath + paradat.jobname + Formtail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")
    paradat.fo = open(paradat.outfile,'w',encoding="utf-8")

    parashow(paradat,paradat.fo)
    ##  parashow(paradat,paradat.listfile)

    ##  read in n-gram data :
    if os.path.isfile(paradat.gramdata):
        gramname = paradat.gramdata
    else:
        gramname = paradat.outpath + paradat.jobname + Gramtail + ".dat"
    print("N-grams to be read from",os.path.realpath(gramname),':')
    paradat.gramfile = open(gramname,'r',encoding="utf-8")
    ng = getgrams(paradat,catinfo)
    print(ng,"n-grams read from",os.path.realpath(paradat.gramfile.name))

    ##  write header info :
    print(paradat.dateline,file=paradat.listfile)
    print("parafile:",paradat.id,file=paradat.listfile)
    print("metafile:",paradat.metafile,file=paradat.listfile)
    print("miniglen:",paradat.miniglen,file=paradat.listfile)
    print("maxiglen:",paradat.maxiglen,file=paradat.listfile)
    print("topgrams:",paradat.topgrams,file=paradat.listfile)
    print(paradat.docs,paradat.cats,file=paradat.listfile)
    print('',file=paradat.listfile)

    ##  identify sequences covered in each category :
    for cn,catname in enumerate(paradat.catlist):
        for dc,d in enumerate(doclist):
            if catname == d.outcome:
                d.marx = grammarx(paradat,catinfo,catname,d)
                ##  sanity check :
                if dc in [-1]:
                    print(d.name)
                    for j,t in enumerate(d.toklist):
                        if d.marx[j]:
                            print(j+1,t,d.marx[j])

    ##  compute coverage by cats & by docs :
    for catname in paradat.catlist:
        catinfo[catname].tokhits = 0
        catinfo[catname].charhits = 0
    sortlist = []
    for cn,catname in enumerate(paradat.catlist):
        for dc,d in enumerate(doclist):
            if catname == d.outcome:
                d.charhits,d.tokhits = covered(paradat,d)
                d.charate = d.charhits / (d.textlen+1)
                d.tokrate = d.tokhits / d.size
                ##  print(dc+1,d.name,d.charate,d.tokrate)
                catinfo[catname].charhits += d.charhits
                catinfo[catname].tokhits += d.tokhits
                sortlist.append((d.charate,d.tokrate,dc))

    catcover(paradat,catinfo)  ##  shows coverage by category
    
    ##  coverage by text :
    print("Document coverage% (characters, tokens) by frequent n-grams :\n",
          file=paradat.listfile)
    sortlist.sort(reverse=True)
    for j,s in enumerate(sortlist):
        cc,tc,dc = s  ##  unpacked
        d = doclist[dc]
        oc = str(d.outcome)
        lout = lineout(d,s)  ##  tuple to printable line
        print(str(j+1).rjust(4)+' '+lout+' '+oc[:12].ljust(12)+' '+d.name,
              file=paradat.listfile)

    ##  extract & sort subsequences :
    extract(paradat,catinfo)
    for c,catname in enumerate(paradat.catlist):
        sortlist = []
        ndox = catinfo[catname].docs
        for t in catinfo[catname].formfrex:
            f = catinfo[catname].formfrex[t]
            nt = len(t)  ##  size in tokens
            st = ' '.join(t)
            x = (len(st)+1) * f  / catinfo[catname].totchars
            ##  character coverage.
            sortlist.append((round(100*x,7),f,nt,t))
        sortlist.sort(reverse=True)
        print(c,catname,ndox,catinfo[catname].tottoks,catinfo[catname].totchars,
              file=paradat.dumpfile)
        jmax = min(len(sortlist),paradat.outforms)  ##  sometimes very few
        fm = "{:7.4f} {:5d} {:4d} {:3d}  {:s}"
        for j in range(jmax):
            v,f,nt,t = sortlist[j]  ##  unpack
            st = ' '.join(t)
            size = len(st)  ##  charsize
            lout = fm.format(v,f,size,nt,st)
            print(lout,file=paradat.dumpfile)

    ##  create flic o/p, if testmeta given :
    tt = os.path.abspath(paradat.testmeta) != os.path.abspath(metaname)
    ##  train & test must differ to make next section worth doing :
    if tt and paradat.testmeta and os.path.isfile(paradat.testmeta):
        paradat1 = paracopy(paradat,paradat.listfile)
        outcomes = metagrab(paradat.testmeta,paradat)
        paradat.outcomes = outcomes
        doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
        paradat.doclist = doclist
        ##  first deal with mixed outcome types :
        paradat.catlist = sorted(list(set(outcomes)))
        paradat.catvals = set(paradat.catlist)
        paradat.cats = paradat.seencats = len(paradat.catvals)
        paradat.docs = len(doclist)
        assert paradat.docs == len(outcomes)
        print("Number of test texts =",len(doclist))
        print("Number of test tokens=",paradat.tottoks)
        print("\n:: Processing files from test metafile "+paradat.testmeta+" :",
              file=paradat1.listfile)
        ##  revise catinfo :
        catinfo1 = catinfo ; catinfo = {}
        for c in paradat.catvals:
            catinfo[c] = Sack(c)
            catinfo[c].tottoks = 0
            catinfo[c].totchars = 0
            catinfo[c].docs = 0
        ##  cumulate sizes by category :
        for d in doclist:
            catinfo[d.outcome].tottoks += d.size
            catinfo[d.outcome].docs += 1
            d.text = ' '.join(d.toklist) + ' '
            d.textlen = len(d.text)  ##  trailing blank included
            catinfo[d.outcome].totchars += d.textlen  ##  total character count
        ##  compute coverage on flic docs :
        sortlist = []
        for catname in paradat.catlist:
            catinfo[catname].tokhits = 0
            catinfo[catname].charhits = 0
        for dc,d in enumerate(paradat.doclist):
            ##  print(dc+1,d.name,d.size,d.textlen,file=paradat1.listfile)
            if d.outcome in paradat1.catvals:
                catname = d.outcome
            else:
                catname = paradat1.refcode
            d.marx = grammarx(paradat,catinfo1,catname,d)
            formchar,formtoks = covered(paradat,d)
            catinfo[d.outcome].charhits += formchar
            catinfo[d.outcome].tokhits += formtoks
            cc = round(formchar/d.textlen,7)
            tc = round(formtoks/d.size,7)
            sortlist.append((cc,tc,dc))
        catcover(paradat,catinfo)  ##  show category coverage
        ##  also document coverage :
        print("Document coverage% (characters, tokens) by frequent n-grams :\n",
              file=paradat.listfile)        
        sortlist.sort(reverse=True)
        for j,s in enumerate(sortlist):
            ##  print(s,file=paradat1.listfile)
            cc,tc,dc = s
            d = doclist[dc]  ##  refers to metflic files
            oc = str(d.outcome)
            lout = lineout(d,s)  ##  tuple to printable line
            print(str(j+1).rjust(4)+' '+lout+' '+oc[:12].ljust(12)+' '+d.name,
                  file=paradat1.listfile)
        print("\nN.B. category",paradat1.refcode,"used if class label previously unseen.",
              file=paradat1.listfile)

    ##  finale :
    paradat.dumpfile.close()
    paradat.listfile.close()
    paradat.fo.close()
    print("listing written onto file",os.path.abspath(paradat.listfile.name))
    print("formdump written onto file",os.path.abspath(paradat.dumpfile.name))    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0




##  CATCOVER :
def catcover (paradat,catinfo):
    "shows n-gram covernage by category."

    print("Category coverage% (characters, tokens) by frequent n-grams :\n",
          file=paradat.listfile)
    for cn,catname in enumerate(paradat.catlist):
        cc = catinfo[catname].charhits / catinfo[catname].totchars
        ##  ??  + .docs
        tc = catinfo[catname].tokhits / catinfo[catname].tottoks
        nout = " {0:8.4f}  {1:8.4f}".format(cc*100,tc*100)
        print(cn,catname.ljust(16)+nout,file=paradat.listfile)
    paradat.listfile.write(NL)

    return


##  LINEOUT :
def lineout (d,s):
    "makes printable line from file-info tuple."

    cc,tc,dc = s
    ##  oc = d.outcome[:12]
    f = "{0:6d} {1:5d} {2:7.2f}{3:7.2f} "
    lout = f.format(d.textlen,d.size,100*cc,100*tc)
    ##  outcome & file name appended outside.

    return  lout


##  EXTRACT :
def extract (paradat,catinfo):
    "extracts & tabulates subsequences in d.marx."

    for c in paradat.catlist:
        catinfo[c].formfrex = {}

    for d in paradat.doclist:
        catname = d.outcome
        d.slabfrex = subseqs(d,paradat.miniglen)
        add2dic(catinfo[catname].formfrex,d.slabfrex)

    return


##  SUBSEQS :
def subseqs (d,consmin=2):
    "returns frequency table of subseqs in d (skipping singletons by default)."

    slabtab = {}
    nt = len(d.toklist)
    d.marx.append(0)  ##  sentinel
    slab = []
    for j,f in enumerate(d.marx):
        if f > 0:  ##  start or continuation of sequence
            slab.append(d.toklist[j])  ##  append corresponding token
            ##  if slab == ["leaflet", "carefully"]:
                ##  print(j,d.name,d.toklist[j-2:j+7])
        else:  ##  sequence ends
            if len(slab) >= consmin:  ##  worth saving
                stup = tuple(slab)
                sf = slabtab.get(stup,0)
                slabtab[stup] = sf + 1
                slab = []  ##  re-start
            else:  ##  too short to save
                slab = []

    del d.marx[-1]  ##  remove sentinel
    return  slabtab


##  COVERED :
def covered (paradat,d):
    "finds coverage by n-grams in 1 document."

    n = len(d.marx)
    assert n == d.size
    ch = th = 0
    for p in range(n):
        if d.marx[p] > 0:  ##  covered by frequent n-grams
            th += 1
            ch += (len(d.toklist[p])+1)  ##  always count (trailing) space

    return  ch,th


##  GRAMMARX :
def grammarx (paradat,catinfo,catname,d):
    "marks locations in d covered by n-grams of type catname."

    ntox = len(d.toklist)
    marx = [0] * ntox
    if catname != d.outcome:
        print(d.name,d.outcome,"treated as from class",catname)
    ##  assert catname == d.outcome
    for item in catinfo[catname].gramlist:
        g = list(item[-1])  ##  n-gram itself, forced 2b a list
        glen = len(g)  ##  can't be zero
        if glen < paradat.miniglen:
            continue  ##  too short
        locs = findlocs(d.toklist,g[0])  ##  find locs of initial token
        for p in locs:
            slab = d.toklist[p:p+glen]
            if slab == g:
                for j in range(p,p+glen):
                    marx[j] += 1  ##  hits at this token position

    return  marx  ##  2b held with d


##  FINDLOCS :
def findlocs (toklist,item):
    "returns locations of item in token list."

    locs = []
    p = 0
    nt = len(toklist)
    while p < nt:
        try:
            p = toklist.index(item,p)
            locs.append(p)
            p += 1
        except ValueError as xx:
            break

    return  locs


##  GETGRAMS :
def getgrams (paradat,catinfo):
    "gets n-grams from gramfile into catinfo sublist."

    datlines = paradat.gramfile.readlines()
    nl = len(datlines)
    print(nl)
    catname = "zonQ"
    ng = 0
    for c in paradat.catlist:
        catinfo[c].gramlist = []  ##  initial gramlist
    for j,datline in enumerate(datlines):
        if not datline:
            continue  ##  skip blanks
        if datline[0] == '#':  ##  category header
            bits = datline.split()
            catname = bits[1]
            if not (catname in paradat.catlist):
                print("Warning, unexpected class name :",catname,"!!")
        elif " (" in datline:
            p = datline.find(" (")
            slab = datline[p:].strip()
            g = eval(slab)  ##  should become a tuple
            if len(g[-1]) >= paradat.miniglen:  ##  last item is n-gram
                ng += 1
                catinfo[catname].gramlist.append(g)  ##  still has size & freq
            ##  else too short 2b worth using
        else:
            pass

    return  ng


##  TOPFREX :
def topfrex (vtab,maxtops):
    "returns high-freq items from vtab dictionary in descending order."

    sortlist = [(vtab[k],k) for k in vtab]
    sortlist.sort(reverse=True)

    return  sortlist[:maxtops]


##  NATS :
def nats (paradat,toktuple):
    "sums info in nats of tokens in toktuple."

    nsum = 0.0
    dnom = paradat.tottoks + paradat.vocsize
    for t in toktuple:
        p = (paradat.allfrex.get(t,0)+1) / dnom
        nsum -= math.log(p)

    return  nsum


##  ALFS :
def alfs (s,p=1):
    "number of alphabetics in a string or sequence."

    a = 0
    if type(s) == type('s'):
        for ch in s:
            if ch.isalpha():
                a += 1
    else:  ##  sequence
        for item in s:
            a += alfs(item)**p

    return  a
        



##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
