"Python3 utility routines for quayness software, by R.S. Forsyth."

##  QUAYLIB :
##
##  Python3 keyness calculation utilities :
##  Copyright 2013, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 26/08/2013
##  last revision : 19/03/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
    This copyrighted program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
	
import  math, os, string, sys
    
from globdefs import *


##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  KEYSHOW :
def keyshow (keydata,fo=sys.stdout):
    "displays keydata for inspection."

    print('',file=fo)
    outpars(keydata,fo)
    for k in keydata.poskeys:
        outpars(k,fo)
    print("==",file=fo)
    for k in keydata.negkeys:
        outpars(k,fo)
    print('',file=fo)

    return


##  GETKEYS :
def getkeys (paradat,doclist,catname,indices):
    "returns Sack with most positive (& negative) keys."

    outstuff = Sack(paradat.metafile)
    outstuff.catname = catname
    outstuff.classnum = paradat.catlist.index(catname)
    outstuff.traindox = len(indices)
    outstuff.keymode = paradat.keymode
    outstuff.poskeys = [] ; outstuff.negkeys = []
    ##  collect candidate tokens :
    docfrex = {}
    for i in indices:
        d = doclist[i]
        tset = set(d.toklist)
        for t in tset:
            if t in docfrex:
                docfrex[t] += 1
            else:  ##  first encounter
                docfrex[t] = 1
    allkeys = []
    nt = 0
    km = paradat.keymode.lower().strip()[:4]  ##  first four chars only
    bins = []
    for t in docfrex:
        if docfrex[t] < paradat.mindocf:
            continue  ##  too rare to bother with
        rawfrex = fournums(paradat,doclist,catname,indices,t)
        f1,n1,f2,n2 = rawfrex
        if km.startswith("coda"):
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = codacalq(inrates,outrates)  ##  same as rocarea() ??
        elif km.startswith("czec"):  ##  czech so-called Dice coeff
            ##  f1,n1,f2,n2 = rawfrex
            kv = czechdif(f1,n2,f2,n2)
        elif km.startswith("doc"):  ##  document-based scores
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            if km == "docz":
                kv = doczeta(inrates,outrates)
            else:
                a,b,c,d = foursome(inrates,outrates)
                slug = math.sqrt(a+c) * 0.5
                a += slug ; c += slug
                slug = math.sqrt(b+d) * 0.5
                b += slug ; d += slug
                if km == "doct":
                    kv = tetracor(a,b,c,d,0.0)
                else:  ##  doch, docs, docx ??
                    kv = phi4(a,b,c,d,0.0)
        elif km.startswith("fazs"):  ##  freq-adjusted z-scores
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = catfazs(inrates,outrates)
        elif km.startswith("g2"):  ##  g-squared
            kv = g2score(f1,n1,f2,n2)
        elif km.startswith("hg") or km.startswith("hedg"):  ##  hedges's g
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = hedgesg(inrates,outrates)  ##  pb corr of rates gives same order.
        elif km.startswith("mz"):  ##  mean-based neozeta
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = meanzeta(inrates,outrates)  ##  * 100.0 ??
        elif km.startswith("neoz"):  ##  snippet-based zeta (-1)
            ix,isum,ox,osum = snipzeta(doclist,indices,catname,t)
            kv = (ix/isum - ox/osum)  ##  merge into 1 function ??
        elif km.startswith("pbcc"):
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            if not bins:
                bins = ([1] * len(inrates)) + ([0] * len(outrates))
            kv = pbcc(inrates,outrates,bins)  ##  same effect as hedg !!
        elif km.startswith("pbcx"):
            xvec,yvec = pbcx(doclist,indices,catname,t)
            cx = corsums(xvec,yvec)
            kv = corrval(cx)
        elif km.startswith("ppqq"):  ##  document version of p*q
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = 100.0 * ppqq(inrates,outrates)
        elif km.startswith("pq"):  ##  snippet version of p*q
            ix,isum,ox,osum = snipzeta(doclist,indices,catname,t)
            f1 = ix / max(1,isum)
            f2 = ox / max(1,osum)
            pq = f1 * (1.0-f2) ; kv = 100.0 * pq
        elif km.startswith("reld"):
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = reldiff(inrates,outrates)
        elif km.startswith("roar"):
            r1 = f1 / max(n1,1) ; r2 = f2 / max(n2,1)
            kv = (1.0-r2) / (1.0-r1)
        elif km in ["snip","snit","phis"]:  ##  snippet-based phiscore or tetracor
            ix,isum,ox,osum = snipzeta(doclist,indices,catname,t)
            ##  kv = (ix-ox)/isum * (1.0-ox/osum)
            ##  kv = (1.0-ox/(osum+math.log1p(osum))) / (1.0-ix/(math.log1p(isum)+isum))
            slug = math.sqrt(isum) * 0.5  ##  separate slugs for each category
            a = ix+slug ; c = (isum-ix) + slug
            slug = math.sqrt(osum) * 0.5
            b = ox+slug ; d = (osum-ox) + slug
            ##  prod = a*d - b*c
            ##  dnom = math.sqrt((a+b)*(c+d)*(a+c)*(b+d))
            ##  kv = prod/dnom
            ##  slug = math.sqrt(isum+osum) * 0.25
            if km == "phis" or km == "snip":
                kv = phi4(a,b,c,d,0.0)
            else:  ##  snit : tetracor
                kv = tetracor(a,b,c,d,0.0)
            ##  if t in ["kyoto","the"]:
                ##  print(t,ix,isum,ox,osum,slug,kv)  ##  phiscore            
        elif km.startswith("tu"):
            inrates,outrates = ratevecs(doclist,indices,catname,t)
            kv = tukiness(inrates,outrates)
        else:  ##  g2 is default (~loglikelihood)
            ##  f1,n1,f2,n2 = rawfrex
            kv = g2score(f1,n1,f2,n2)
        r1 = f1 / max(n1,1) ; r2 = f2 / max(n2,1)  ##  secondary sortkey
        allkeys.append((kv,r1-r2,docfrex[t],rawfrex,t))
        nt += 1
    allkeys.sort(reverse=True)
    for j in range(paradat.topkeys):
        if j < nt:
            kp = Sack(j)
            kp.quayval,kp.ratediff,kp.docfreq,kp.rawfrex,kp.item = allkeys[j]
            kp.r1 = kp.rawfrex[0] / kp.rawfrex[1]
            kp.r2 = kp.rawfrex[2] / kp.rawfrex[3]
            inrates,outrates = ratevecs(doclist,indices,catname,kp.item)
            kp.avrate1 = sum(inrates) / len(inrates)
            kp.avrate2 = sum(outrates) / len(outrates)
            kp.midrate1 = midvalue(inrates)
            kp.midrate2 = midvalue(outrates)
            ratevex = inrates+outrates
            bs = bastats(ratevex) ; kp.sd = bs.std2
            kp.midrate = midvalue(ratevex)
            outstuff.poskeys.append(kp)
        jj = nt - (j+1)
        if jj >= 0:
            kn = Sack(jj)
            kn.quayval,kn.ratediff,kn.docfreq,kn.rawfrex,kn.item = allkeys[jj]
            kn.r1 = kn.rawfrex[0] / kn.rawfrex[1]
            kn.r2 = kn.rawfrex[2] / kn.rawfrex[3]
            inrates,outrates = ratevecs(doclist,indices,catname,kn.item)
            kn.avrate1 = sum(inrates) / len(inrates)
            kn.avrate2 = sum(outrates) / len(outrates)
            kn.midrate1 = midvalue(inrates)
            kn.midrate2 = midvalue(outrates)
            ratevex = inrates + outrates
            bs = bastats(ratevex) ; kn.sd = bs.std2
            kn.midrate = midvalue(ratevex)
            outstuff.negkeys.append(kn)
    outstuff.negkeys.reverse()  ##  put into descending order
        
    return  outstuff


##  RATEVECS :
def ratevecs (doclist,indices,thiscat,t):
    "returns rates of t separately within & outside thiscat."

    outrates = []
    inrates = []
    for j in indices:
        d = doclist[j]
        dcat = d.outcome
        f = d.freqtab.get(t,0)
        r = f / d.size
        if dcat == thiscat:
            inrates.append(r)
        else:
            outrates.append(r)

    return  (inrates,outrates)


##  PPQQ :
def ppqq (inrates,outrates):
    "p*q."

    z1 = inrates.count(0.0)
    z2 = outrates.count(0.0)
    p = 1.0 - z1/len(inrates)
    q = z2/len(outrates)
    ##  zero-counts, so inverted.

    return  p*q


##  PBCC :
def pbcc (inrates,outrates,bins):
    "point-biserial correlation coefficient."

    xvec = inrates + outrates
    cx = corsums(xvec,bins)  ##  bins is 1/0 index vector

    return  corrval(cx)  ##  just r with 1 vector binary
    ##  just gives same result as hedges's g !!


##  PBCX :
def pbcx (doclist,indices,thiscat,t):
    "revised pbcc."
    
    xvec = []
    yvec = []
    for j in indices:
        d = doclist[j]
        dcat = d.outcome
        f = d.freqtab.get(t,0)
        r = f / d.size
        xvec.append(r**0.5)  ##  reduce positive skew
        if dcat == thiscat:
            yvec.append(d.size**0.5)  ##  weighted by sqrt(size)
        else:
            yvec.append(-d.size**0.5)

    return  (xvec,yvec)


##  CZECHDIF :
def czechdif (f1,n1,f2,n2):
    "computes czech version of 'Dice' coefficient."

    if n1 < 1 or n2 < 1:
        return  0.0
    r1 = (f1+1)/(n1+2) ; r2 = (f2+1)/(n2+2)

    return  (r1 - r2) / (r1 + r2)


##  RELDIFF :
def reldiff (inrates,outrates):
    "simple difference in mean rates."

    n1 = len(inrates) ; n2 = len(outrates)
    if n1 < 1 or n2 < 1:
        return  0.0
    m1 = sum(inrates) / n1
    m2 = sum(outrates) / n2
    if m1 <= 0.0 and m2 <= 0.0:
        return  0.0  ##  should be impossible

    return  (m1 - m2) * 100.0  ##  percentage
    

##  CATFAZS :
def catfazs (inrates,outrates):
    "categoric version of fazs."

    n1 = len(inrates) ; n2 = len(outrates)
    p = n1 / (n1+n2)
    s = (p*(1.0-p)) ** 0.5  ##  sd using binomial
    z1 = (1.0-p) / s ; z2 = -p / s
    totprod = 0.0
    for r in inrates:
        totprod += z1 * r
    for r in outrates:
        totprod += z2 * r

    return  100.0*totprod / (n1+n2)


##  MIDZETA :
def midzeta (inrates,outrates):
    "computes document-based Burrows/Craig zeta - 1."
    
    ri = midvalue(inrates)
    ro = midvalue(outrates)
            
    return  ri - ro


##  MEANZETA :
def meanzeta (inrates,outrates):
    "computes (heavily!) modified Burrows/Craig zeta - 1."
    
    ri = sum(inrates) / max(1,len(inrates))
    ro = sum(outrates) / len(outrates)
    mz = ri - ro
            
    return  mz  ##  equivalent to reldiff (& fazs!)


##  DOCZETA :
def doczeta (inrates,outrates):
    "computes (heavily!) modified Burrows/Craig zeta - 1."

    n2 = len(outrates)
    if n2 < 1:
        return  0.0
    z1 = inrates.count(0.0)
    z2 = outrates.count(0.0)
    n1 = len(inrates)
    docz = (n1-z1) / max(1,n1) - (n2-z2) / n2
            
    return  docz


##  SNIPZETA :
def snipzeta (doclist,indices,thiscat,t):
    "returns snippet rates of t separately within & outside thiscat."

    intotal = outtotal = 0
    isum = osum = 0
    for j in indices:
        d = doclist[j]
        f = 0
        for s in d.snipsets:
            if t in s:
                f += 1  ##  number of snippets in which t is found
        if d.outcome == thiscat:
            intotal += f
            isum += d.snippets
        else:
            outtotal += f
            osum += d.snippets

    return  (intotal,isum,outtotal,osum)
    ##  worth precomputing sniptab for each document ??


##  CODACALC :
def codacalc (inrates,outrates):
    "returns coefficient of differential abundance."

    hi = lo = 0
    n1 = len(inrates)
    n2 = len(outrates)
    nn = n1 * n2
    for r1 in inrates:
        for r2 in outrates:
            if r1 > r2:
                hi += 1
            elif r2 > r1:
                lo += 1
    eq = nn - (hi+lo)
    ##  coda = (hi - lo) / nn
    coda = (hi + 0.5*eq) / nn
    
    return  coda


##  CODACALQ :
def codacalq (inrates,outrates):
    "coefficient of differential abundance, only with nonnegative values."

    hi = 0
    n1 = len(inrates)
    n2 = len(outrates)
    ovec = sorted(outrates)
    ##  z1 = inrates.count(0.0)
    ##  print(inrates) ; print(ovec)
    z2 = ovec.count(0.0)  ##  counts integer 0 as well
    eq = 0
    nn = n1 * n2
    for r1 in inrates:
        if r1 == 0.0:  ##  lo += (n2-z2)
            eq += z2
        else:
            hi += z2
            for j in range(z2,n2):  ##  look at nonzeroes
                r2 = ovec[j]
                if r1 > r2:
                    hi += 1
                elif r1 == r2:
                    eq += 1
                else:  ## r1 < r2:
                    break  ##  will lose from here on

    coda = (hi + 0.5*eq) / nn
    
    return  coda
    ## should be quick when given lots of zeroes (which is common).


##  CODACALX :
def codacalx (inrates,outrates):
    "supposedly faster version of codacalc."

    hi = eq = 0
    n1 = len(inrates) ; n2 = len(outrates)
    nn = n1 * n2
    ivec = sorted(inrates)
    ovec = sorted(outrates)
    ##  print(ivec) ; print(ovec)
    halfway = n1 // 2
    for r1 in ivec[:halfway]:
        ec = 0
        for j,r2 in enumerate(ovec):
            if r1 == r2:
                ec += 1
            elif r1 < r2:
                break
        ##  j points to first loss.
        wins = j - ec
        ##  print(r1,wins,ec)
        hi += wins ; eq += ec
    for r1 in ivec[halfway:]:
        ec = 0
        for j in range(n2-1,-1,-1):
            r2 = ovec[j]
            ##  print("!!",r1,j,r2)
            if r1 == r2:
                ec += 1
            elif r1 > r2:
                break
        ##  j points to first win.
        wins = (j+1)
        hi += wins ; eq += ec
        ##  print(r1,wins,ec)
    coda = (hi + 0.5*eq) / nn
    
    return  coda
    ##  slower !!


##  CODACALK :
def codacalk (inrates,outrates):
    "supposedly faster version of codacalc, using binary chop."

    hi = eq = 0
    n1 = len(inrates) ; n2 = len(outrates)
    nn = n1 * n2
    ##  ivec = sorted(inrates)
    ovec = sorted(outrates)
    ##  print(inrates) ; print(ovec)
    for r1 in inrates:
        lh = 0 ; rh = n2 - 1  ##  edges
        test = False
        while (rh >= lh) and not test:
            midpoint = (lh+rh) // 2
            if r1 > ovec[midpoint]:  ##  go up
                lh = midpoint + 1
            elif r1 < ovec[midpoint]:  ##  go down
                rh = midpoint - 1
            else:
                test = True
        if test:  ##  deal with equals
            ##  print(r1,lh,midpoint,rh)
            m = midpoint
            while m < n2:
                if ovec[m] == r1:
                    eq += 1
                else:
                    break
                m += 1
            m = midpoint - 1
            while m >= 0:
                if ovec[m] == r1:
                    eq += 1
                else:
                    break
                m -= 1
            hi += (m + 1)
        else:
            ##  print(test,r1,lh,midpoint,rh)
            hi += lh
            ##  eq += 0

    coda = (hi + 0.5*eq) / nn
    return  coda  ##  actually slower! (many equalities?)
    

##  TUKINESS :
def tukiness (inrates,outrates):
    "robust keyness index based on five-number reference levels."

    ##  fails if less than 2 in each vector !
    n1 = len(inrates) ; n2 = len(outrates)
    if n1 < 1 or n2 < 1:
        print(n1,n2)
        raise RuntimeError("tukiness vectors too short!!")
    y5 = fivenum(outrates)
    while len(y5) < 5:
        m = sum(y5) / len(y5)
        y5.append(m)  ##  kludge to deal with short vectors

    return  superior(inrates,y5)


##  TUQI :
def tuqi (inrates,outrates):
    "robust keyness index based on five-number reference levels."

    ##  fails if less than 2 in each vector !
    n1 = len(inrates) ; n2 = len(outrates)
    if n1 < 1 or n2 < 1:
        print(n1,n2)
        raise RuntimeError("tuqi vectors too short!!")
    x5 = fivenum(inrates)
    m1 = sum(inrates) / n1
    z1 = inrates.count(0.0)
    while len(x5) < 6:
        x5.append(m1)  ##  quasi-kludge to deal with short vectors
    y5 = fivenum(outrates)
    m2 = sum(outrates) / n2
    z2 = outrates.count(0.0)
    while len(y5) < 6:
        y5.append(m2)  ##  quasi-kludge also ensures mean is present
    hits = 0
    for x in x5:  ##  really 6!
        for y in y5:
            if x>y:  hits += 2
            elif x==y:  hits += 1
    ##  r = ((n1-z1)+z2)/(n1+n2)
    ##  r = ((n1-z1)/n1) * (z2/n2)
    r = m1 - m2
    return  hits + r  ##  / 72
    ##  hommage to J.Tukey.


##  FOURNUMS :
def fournums (paradat,doclist,catname,indices,t):
    "counts occurrences of token t within & outside given class."

    f1 = n1 = 0
    f2 = n2 = 0
    for i in indices:
        d = doclist[i]
        ##  tf = d.toklist.count(t)  ##  very slow !!
        tf = d.freqtab.get(t,0)
        if d.outcome == catname:  ##  selected class
            f1 += tf
            n1 += d.size
        else:  ##  unselected class
            f2 += tf
            n2 += d.size

    return  (f1,n1,f2,n2)


##  G2SCORE :
def g2score (f1,n1,f2,n2):
    "computes g-squared score from computed frequencies."

    if n1 < 1 or n2 < 1:
        return  0.0
    gx = g2(f1,n1,f2,n2)
    gval = gx[0]
    if gx[1] <= 0.0:
        gval = -gval

    return  gval  ##  signed value

        
##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  OUTPARS :
def outpars (paradat,listing=sys.stdout):
    "writes printable parameters to file as o/p header."

    parlist = sorted(dir(paradat))
    for p in parlist:
        if p.startswith("__"):  ##  private
            continue
        v = getattr(paradat,p)
        pout = ''
        if isinstance(v,int):
            pout = str(v)
        elif isinstance(v,float):
            pout = str(round(v,5))
        elif isinstance(v,str):
            pout = v
        elif isinstance(v,(dict,list,set,tuple)):
            if len(v) < 10:
                pout = str(v)
            else:
                tn = str(type(v))
                p1 = tn.find("'")
                p2 = tn.rfind("'")
                pout = tn[p1:p2+1]+" of "+str(len(v))+" items."
        else:
            fd = dir(v)
            if "mode" in fd and ("name" in fd):  ##  must be a file ?
                pout = "~ " + v.name
        if pout:
            print(p.ljust(9),pout,file=listing)

    print('',file=listing)
    return


##  ADDSNIPS :
def addsnips (paradat,doclist):
    "attaches snippet info to doclist."

    totsnips = 0
    minsnip = int(round(paradat.snipsize / Gold))
    for d in doclist:
        d.snippets = 0
        d.snipsets = []  ##  list of sets
        for p in range(0,d.size,paradat.snipsize):
            slab = d.toklist[p:p+paradat.snipsize]
            if len(slab) >= minsnip:
                d.snipsets.append(set(slab))
                d.snippets += 1
                totsnips += 1

    return  totsnips
    ##  later have special case for small docs ??


##  Stat routines :

##  BASTATS :
def bastats (numlist):
    "basic stats routine for a list of numbers."

    fn = float(len(numlist))
    sum1 = 0.0
    sum2 = 0.0
    for x in numlist:
        sum1 += x
        sum2 += x**2

    statsack = Sack(0)
    ##  defaults for dodgy inputs :
    statsack.mean = sum1
    statsack.var1 = statsack.var2 = sum2
    statsack.std1 = statsack.std2 = math.sqrt(abs(sum2))
    if fn > 0.0:
        statsack.mean = sum1 / fn
        statsack.var1 = sum2 / fn - statsack.mean**2
        ##  print "$$",statsack.var1
        statsack.std1 = math.sqrt(max(0.0,statsack.var1))
        if fn > 1.0:
            statsack.var2 = statsack.var1 * (fn/(fn-1.0))
            statsack.std2 = math.sqrt(max(0.0,statsack.var2))

    return statsack


##  RANX :
def ranx (xvec):
    "returns rank vector from raw data vector, with ties given mean positions."

    nx = len(xvec)
    ##  xvec.sort()  ##  ascending
    vtab = {}
    for x in xvec:
        fx = vtab.get(x,0)
        vtab[x] = fx + 1
    vvec = list(vtab.items())
    vvec.sort()
    ranktab = {}
    p = 0
    ranksum = 0
    for v in vvec:
        vx = v[0] ; vf = v[1]
        rx = (ranksum*2+vf)*0.5-0.5  ##  plus 0.5 if lowest rank is 1
        ranksum += vf
        ##  print(p,vx,vf,rx)
        ranktab[vx] = rx
        p += 1

    rankvec = []
    for x in xvec:
        rankvec.append(ranktab[x])
    return  rankvec  ##  returning ranktab more useful ?


##  CORSUMS :
def corsums (x,y):
    "produces sums required for correlation coefficient, in a Sack()."

    cs = Sack(0)
    n = len(x)
    n = min(n,len(y))  ##  chops extra if unequal
    if n < 1:
        print("Problem in corsums!!")
        return cs
    sx = sy = 0.0
    sxsq = sysq = 0.0
    sumprods = 0.0
    for i in range(n):
        sx += x[i] ; sy += y[i]
        sxsq += x[i]**2 ; sysq += y[i]**2
        sumprods += x[i] * y[i]

    cs.n = float(n)
    cs.sumprods = sumprods
    cs.sx = sx
    cs.sy = sy
    cs.sxsq = sxsq
    cs.sysq = sysq
    return  cs


##  CORRVAL :
def corrval (cs,epsilon=1e-23):
    "computes correlation coefficient from sums etc. in Sack()."

    cnum = cs.sumprods - (cs.sx*cs.sy) / cs.n
    tx = cs.sxsq - cs.sx**2 / cs.n
    ty = cs.sysq - cs.sy**2 / cs.n
    ##  try:
        ##  dnom = math.sqrt(tx*ty)
    ##  except ValueError as xxxx:
        ##  print(xxxx)
        ##  print(tx,ty,cnum,cs.n)
        ##  print(cs.sx,cs.sy,cs.sxsq,cs.sysq)
        ##  dnom = 0
    ##  dnom = math.sqrt(abs(tx*ty))  ##  danger of rounding errors
    dnom = math.sqrt(max(0.0,tx*ty))
    if dnom > epsilon:
        return  cnum / dnom
    elif tx < epsilon and ty < epsilon:
        return  1.0  ##  both flat
    else:
        return  0.0  ##  null case
    ##  semi-safe, but vulnerable to rounding errors.


##  FIVENUM :
def fivenum (xvec):
    "Tukey-style five-number summary of numeric vector."

    n1 = len(xvec) - 1
    xs = sorted(xvec)
    if n1 < 5:
        return xs
    v5 = [xs[0]]  ##  minimum
    for f in [0.25,0.5,0.75]:
        lo = int(n1*f)
        hi = n1 - int((1.0-f)*n1)
        v5.append((xs[lo]+xs[hi])*0.5)
    v5.append(xs[-1])  ##  maximum

    return  v5


##  SUPERIOR :
def superior (xvec,y5):  ##  y5 must be fivenum of comparison vector
    "nonparametric fuzzy vectorial greater-than."

    nx = len(xvec)
    if len(y5) != 5:
        raise RuntimeError("y5 must be 5num summary!!")
    wins = 0
    for x in xvec:
        for y in y5:
            if x > y:
                wins += 2
            elif x == y:
                wins += 1
    
    return  wins / (10.0 * nx)


##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5


##  G2 :
def g2 (f1,n1,f2,n2):
    "special case of log-likelihood function for corpus frequency comparison."

    totfreq = f1 + f2
    grandtot = n1 + n2  ##  typically n1 is smaller, though not essential
    if totfreq < 4:
        return  (0.0,0,0)
    rate = totfreq / float(grandtot)
    e1 = rate * n1 ; e2 = rate * n2
    ##  print f1,n1,e1,"*",f2,n2,e2
    if e1 > Tiny and f1 > 0:
        x1 = f1 * math.log(f1/e1)
    else:
        x1 = 0.0
    if e2 > Tiny and f2 > 0:
        x2 = f2 * math.log(f2/e2)
    else:
        x2 = 0.0

    return (2*(x1+x2),x1,x2)  ##  returns a triple, item zero is g2
    ##  sign of x1 is directional indicator, x1 <= 0 indicates negative assoc.


##  HEDGESG :
def hedgesg (x1,x2):
    "effect-size measure for 2 independent groups: Cohen's d with pooled variance."

    n1 = len(x1) ; n2 = len(x2)
    if (n1+n2) < 4 or n1 < 1 or n2 < 1:
        return  0.0
    s1 = bastats(x1)
    s2 = bastats(x2)
    diff = s1.mean - s2.mean
    if (n1+n2) <= 4:  ##  puny sample
        print("warning, small samples in hedgesg() !!")
        dnom = max(s1.std2,s2.std2)
        if dnum <= 0.0:
            dnom = 1.0
        return  diff / dnom
    else:  ##  normal case
        pooledv = ((n1-1)*s1.var2 + (n2-1)*s2.var2) / (n1+n2-2)
        if pooledv <= 0.0:
            pooledv = 1.0  ##  should give warning ?
        return  diff / math.sqrt(pooledv)


##  TETRACOR :
def tetracor (a,b,c,d,slug=0.5):
    "tetrachoric correlation coefficient from fourfold table counts [11,10;01,00]."

    x = math.sqrt((a+slug)*(d+slug))  ##  attenuation factor added
    y = math.sqrt((b+slug)*(c+slug))
    n = x - y
    d = x + y + Tiny
    return  math.sin(0.5*math.pi*(n/d))


##  TET4 :
def tet4 (a,b,c,d,slug=0.5) :
    "alternative formula for estimating tetrachoric correlation."

    x = (a+slug)*(d+slug)  ##  attenuation factor added
    y = (b+slug)*(c+slug)  ##  slug must not be negative
    d = 1.0 + math.sqrt(x/max(y,Tiny))
    ##  print(x,y,d)
    ##  less stable with small numbers.

    return  math.cos(math.pi / max(d,Tiny))


##  YULEVALQ :
def yulevalq (a,b,c,d,slug=0.5):
    "Yule's Q association coefficient from fourfold table counts [11,10;01,00]."

    x = (a+slug)*(d+slug)  ##  attenuation factor added
    y = (b+slug)*(c+slug)  ##  slug must not be negative
    n = x - y
    d = x + y + Tiny
    return  n/d


##  YULEVALY :
def yulevaly (a,b,c,d,slug=0.5):
    "Yule's Y colligation coefficient from fourfold table counts [11,10;01,00]."

    x = (a+slug)*(d+slug)  ##  attenuation factor added
    y = (b+slug)*(c+slug)  ##  slug must not be negative
    if x+y <= 1.0:
        return  0.0
    oddz = y / max(Tiny,x)

    return  (1-oddz**0.5) / (1+oddz**0.5)


##  PHI4 :
def phi4 (a,b,c,d,slug=0.5):
    "phi coefficient from fourfold table cell values."

    x = (a+slug)*(d+slug)  ##  attenuation factor added
    y = (b+slug)*(c+slug)  ##  slug must not be negative
    a += slug ; b += slug ; c += slug ; d += slug
    n = a*d - b*c
    dsqd = (a+b)*(a+c)*(b+d)*(c+d)

    return  n / max(Tiny,dsqd**0.5)


##  PHILITE :
def philite (a,b,c,d,slug=0.5):
    "adjusted phi coefficient from fourfold table cell values."

    if d > (a+b+c):  ##  double negative shouldn't dominate (e.g. in corpora)
        d = d ** 0.5  ##  assumes d is no-no cell : reduced in importance
    x = (a+slug)*(d+slug)  ##  attenuation factor added
    y = (b+slug)*(c+slug)  ##  slug must not be negative
    a += slug ; b += slug ; c += slug ; d += slug
    n = a*d - b*c
    dsqd = (a+b)*(a+c)*(b+d)*(c+d)

    return  n / max(Tiny,dsqd**0.5)


##  FOURSOME :
def foursome (xvec,yvec,m=None):
    "makes fourfold frequencies with respect to threshold value."

    ##  in none given, medians used :
    if m == None:
        xm = midvalue(xvec)
        ym = midvalue(yvec)
    m = (xm + ym) * 0.5
    a = b = c = d = 0
    for x in xvec:
        if x > m:
            a += 1  ##  tp
        else:
            c += 1  ##  fp
    for y in yvec:
        if y > m:
            b += 1  ##  fn
        else:
            d += 1  ##  tn

    return  (a,b,c,d)

##  SUBTUP0 :
def subtup0 (g,t):
    "is g a subsequence of t? (long-winded version)"

    ng = len(g)
    if ng < 1:
        return  False  ##  empty g doesn't count as subsequence
    nt = len(t)
    if nt < ng:
        return  False
    locs = []
    p = 0
    while p < len(t):
        try:
            tp = t.index(g[0],p)
            p = tp + 1
            locs.append(tp)
            ##  print(tp,p,locs)
        except ValueError:
            break
    ##  print(locs)
    for tp in locs:
        y = (g == t[tp:tp+ng])
        if y:
            return  True
        ##  print(tp,g,t[tp:tp+ng],y)

    return  False


##  SUBTUPLE :
def subtuple (g,t):
    "is g a subsequence of t? (also works with lists.)"

    ng = len(g)
    if ng < 1:
        return  False  ##  empty g doesn't count as subsequence
    nt = len(t)
    if nt < ng:
        return  False
    p = 0
    while p < len(t):
        try:
            tp = t.index(g[0],p)
            if g == t[tp:tp+ng]:
                return  True
            else:
                p = tp + 1  ##  next occurrence of leader ?
        except ValueError:
            break

    return  False
    ##  slower but more flexible than using join & strings ?
    ##  N.B. g & t must both be lists or both tuples; mixing fails !


##  SCKAPPA2 :
def sckappa2 (ctab,codelist):
    "Siegel & Catellan's version of Kappa, for 2-way confusion matrix held as table."

    nc = len(codelist)
    chap = 0.0  ##  proportion agreement expected by chance
    totf = 0.0  ##  total frequency, all cells in table
    hits = 0
    codefrex = {}
    for k1 in codelist:
        ##  codelist holds all valid category codes.
        for k2 in codelist:
            pair = (k1, k2)
            f = ctab.get(pair,0)
            if k1 == k2:
                hits += f
            f1 = codefrex.get(k1,0)
            codefrex[k1] = f1 + f
            f2 = codefrex.get(k2,0)
            codefrex[k2] = f2 + f
            totf = totf + f
    totf = max(Tiny,totf)  ##  to prevent zero division
    for k in codelist:
        f = codefrex.get(k,0)
        pk = f / (2.0 * totf)
        chap += (pk * pk)
        ##  print k,f,pk,chap
    ##  chap is sum of squared probs.
    pa = hits / totf
    ##  print  hits, pa, totf
    k = (pa - chap) / max((1 - chap),Tiny)
    return  (k, chap,hits,pa,totf)




if 0 > 0:
    zonk = (9,0,9,1,2,3,4,5,9,0,99,5,6)
    s = (9,0,99,6)
    ##  s = zonk[4:]
    ##  zonk = [0,5,6,7]
    print(zonk)
    for j in range(len(s)):
        z = s[:j+1]
        print(j,z); print(subtuple(z,zonk))


if 0 > 0:
    x = [0,0.5,1,4,3,7,0,2.2,20,1.25,0.0]
    y = [4,7,10,8,0,1,2,4,9,12,4,1.25,0.0]
    c = codacalc(x,y)
    cx = codacalx(x,y)
    cz = codacalk(x,y)
    cq = codacalq(x,y)
    print(c,cx,cz,cq)
