"Python3 multi-item term finder."

##  TERMITE :
##
##  Text-Exploring Ranked Multi-Item Term Extractor :
##  Copyright 2013, R.S. Forsyth.
##  first version : 02/09/2013
##  last revision : 26/03/2015


import  math, os, string, sys
import  random, time


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  add2dic
from quaylib import  bastats, ranx
##from statlib import  corsums, corrval
from quaylib import  fivenum, midvalue
from quaylib import  addsnips
from quaylib import  getkeys, keyshow
from quaylib import  getfrex
from quaylib import  outpars


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","termite"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfanum initial
    ("mindocf",3,0,10),  ##  minimum worthwhile document frequency
    ("topkeys",32,1,256),  ##  number of high-scoring keys to keep
    ("snipsize",115,4,2048),  ##  number of tokens in a snippet (57=ssss)
    ("sidespan",5,1,9),  ##  maximum tokens either side of seed
    ("keymode","g2"),  ##  choice of keyness function :
    ##  coda, czec, doczeta, g2, hedg, neozeta, reld, roar, (tuki).
    ("metafile",''),  ##  training corpus
    ("trainmet",''),  ##  ditto (alternative)
    ("dumpfile",''),  ##  machine-readable results file
    ("termfile",''),  ##  human-readable results file (desmoglyphs)
    ##  ("keyfile",''),  ##  file for input of keys (unused!!)
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"termite.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "3.1"
Gold = 5.0**0.5*0.5+0.5
Mindocs = 4
Dumptail = "_dump"
Termtail = "_term"
Keytail = "_keys"
Longname = 28  ##  longest class name; longer will be truncated
Maxtoks = 16 * 1024  ##  will only seek tokens in front chunk of text file


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if not paradat.trainmet:
        paradat.trainmet = paradat.metafile
    elif not paradat.metafile:
        paradat.metafile = paradat.trainmet
    assert  paradat.metafile == paradat.trainmet
    if paradat.keymode.startswith("hg"):
        paradat.keymode = "hedg"  ##  standard form of Hedges's g
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of main (training) corpus :
    metaname = getattr(paradat,"metafile",'')
    if not metaname:
        print("Can't proceed without metafile!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
        if paradat.cats < 2:
            print("Must have at least 2 categories !!")
            print("Program halts.")
            return  -2
    else:
        print("No. of text types =",paradat.cats)

    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    paradat.midsize = midvalue(paradat.sizelist)
    paradat.minsize = min(paradat.sizelist)
    print("Shortest =",paradat.minsize,"tokens.")
    print("median size =",paradat.midsize)
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("mean size =",round(paradat.meansize,2))
    print("Longest =",paradat.maxsize,"tokens.")
    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -5
    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].tottoks)
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  reference class is largest
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)
    print("reference category : ",paradat.refcode)

    for c in paradat.catlist:
        print(c,catinfo[c].docs)

    ##  attach extra attributes to doclist :
    allfrex = {}
    paradat.totsnips = addsnips(paradat,doclist)
    for d in doclist:
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  attach category number as well as class name.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)  ##  each text has freqtab attached
        d.vocsize = len(d.freqtab)
        ##  also make overall freqtabs :
        add2dic(allfrex,d.freqtab)
        ##  need to add into catinfo ??
        d.text = ' '.join(d.toklist) + ' '
        d.textlen = len(d.text)
    print("snippets =",paradat.totsnips)
    paradat.doclist = doclist  ##  copy reference
    paradat.allfrex = allfrex  ##  ditto
    paradat.vocsize = len(allfrex)
    for j in []:  ##  sanity check
        d = doclist[j]
        print(d.dnum,d.name,d.outcome,d.size,d.vocsize,d.snippets)
        for j in [0,-1]:
            s = d.snipsets[j]
            print(len(s),sorted(s))

    ##  prepare o/p files :
    if paradat.dumpfile:
        dumpname = paradat.dumpfile
    else:
        dumpname = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    paradat.dumpfile = open(dumpname,'w',encoding="utf8")
    ##  also need separate results file :
    if paradat.termfile:
        listfile = paradat.termfile
    else:
        listfile = paradat.outpath + paradat.jobname + Termtail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")

    ##  initial checking :
    ##  outpars(paradat)
    ##  print(type(paradat))  ##  needs str() to make it a string
    ##  print(isinstance(paradat,type(Sack())))  ##  => True
    indices = list(range(paradat.docs))  ##  all indices

    ##  find keys :
    keydata = []
    print("extracting keys :")
    for catname in paradat.catlist:
        print(catname)
        keydata.append(getkeys(paradat,doclist,catname,indices))
    outpars(paradat,paradat.dumpfile)
    ##  later read from file (to allow manual inserts) ??

    ##  find multi-item terms :
    for cn,catname in enumerate(paradat.catlist):
        termlist = []
        print(cn,catname)
        for pk in keydata[cn].poskeys:
            ##  print(pk.item,pk.rawfrex)
            tt = termext(paradat,pk.item,doclist,paradat.sidespan,catname)
            for tx in tt:
                ##  print("=>",tx)
                termlist.append(tx)
        print(NL,cn,catname,file=paradat.dumpfile)
        termlist.sort(reverse=True)
        catinfo[catname].termlist = termlist
        for tt in termlist:
            print(tt,file=paradat.dumpfile)
        print(NL,file=paradat.dumpfile)
    
    ##  show results :
    print(paradat.dateline,file=paradat.listfile)
    print("parafile:",paradat.id,file=paradat.listfile)
    print("metafile:",paradat.metafile,file=paradat.listfile)
    print("keymode :",paradat.keymode,file=paradat.listfile)
    print("sidespan:",paradat.sidespan,file=paradat.listfile)
    print(paradat.docs,paradat.cats,file=paradat.listfile)
    print('',file=paradat.listfile)
    print("Desmoglyphs by category :",
            file=paradat.listfile)
    for cn,catname in enumerate(paradat.catlist):
        outterms(paradat,cn,catname,catinfo)
        
    
    ##  finale :
    paradat.dumpfile.close()
    paradat.listfile.close()
    print("desmoglyphs written onto file",os.path.abspath(paradat.listfile.name))
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0




##  OUTTERMS :
def outterms (paradat,cn,catname,catinfo):
    "writes results of term-finding onto already-open listfile."

    print('\n#',cn,catname,file=paradat.listfile)
    headline = " Sortval Natscore docrate+ docrate-"+"seed".rjust(19)+"    subsequence"
    print(headline,file=paradat.listfile)
    ##  sort again :
    seqlist = []
    for tx in catinfo[catname].termlist:
        ##  unpack term-tuple :
        v,t,term,v1,v2 = tx  ##  fragile !!
        if len(term) < 2:
            break  ##  singletons don't count ??
        ts = ' '.join(term)+ ' ' ##  string version
        ##  collect stats about t & ts :
        vec1,vec2 = seqfrex(paradat,cn,t,ts)
        pp = vec1[2]/(vec1[2]+vec1[3])
        ##  fm = "{0:8.2f} {1:7.2f}% {2:7.2f}%   {3:>16s} {4:2s} {5:26s}"
        p2 = vec2[2]/(vec2[2]+vec2[3])
        qq = 1-p2
        sval = v*pp*qq
        seqlist.append((sval,tx,vec1,vec2))
    seqlist.sort(reverse=True)
    seqs = set()
    for desglyph in seqlist:
        ##  unpack term-tuple :
        sval,tx,vec1,vec2 = desglyph
        v,t,term,v1,v2 = tx  ##  fragile !!
        if len(term) < 2:
            break  ##  singletons don't count ??
        ts = ' '.join(term) + ' '  ##  string version
        seen = False
        for s in seqs:
            if ts in s:
                seen = True
                break
        if seen:
            continue
        else:
            link = "=>"
            seqs.add(ts)
        ##  lout = "{0:8.2f}  {1:16s} => {2:26s}".format(v,t,ts)
        ##  collect stats about t & ts :
        ##  vec1,vec2 = seqfrex(paradat,cn,t,ts)
        ##  print(lout,file=paradat.listfile)
        ##  p1 = 100*vec1[0]/(vec1[0]+vec1[1])
        pp = 100*vec1[2]/(vec1[2]+vec1[3])
        f0 = "{0:8.2f}".format(sval)
        fm = " {0:8.2f} {1:7.2f}% {2:7.2f}%   {3:>16s} {4:2s} {5:25s}"
        p2 = 100*vec2[2]/(vec2[2]+vec2[3])
        if pp > p2:  ##  suppress terms with inverted docfrex
            print(f0+fm.format(v,pp,p2,t,link,ts),file=paradat.listfile)

    return


##  SEQFREX :
def seqfrex (paradat,cn,t,ts):
    "finds document frequencies of t & ts in text categories."

    vec1 = [0,0,0,0]
    vec2 = [0,0,0,0]
    n1 = n2 = 0
    catname = paradat.catlist[cn]
    for d in paradat.doclist:
        ##  assert d.text[-1] == ' '
        if d.outcome == catname:
            n1 += 1
            if t in d.text:
                vec1[0] += 1
            else:
                vec1[1] += 1
            if ts in d.text:
                vec1[2] += 1
            else:
                vec1[3] += 1
        else:
            if t in d.text:
                vec2[0] += 1
            else:
                vec2[1] += 1
            if ts in d.text:
                vec2[2] += 1
            else:
                vec2[3] += 1

    return  vec1,vec2


##  TERMEXT :
def termext (paradat,t,doclist,slabsize=4,catname='',cutvalue=1.0):
    "extends terms with frequent predecessors &/or followers."

    tf = 0  ##  will recompute token frequency
    predtab = {} ; succtab = {}
    ##  tabulate predecessors & successors :
    for d in doclist:
        if catname and (d.outcome != catname) or (d.size < 4*slabsize):
            continue  ##  wrong class or too small
        if t in d.freqtab:
            ##  decide how many probes :
            blox = 3 - (d.size<144) + (d.size>=1024) + (d.size>=4096)
            ##  compute bloc boundaries :
            bvec = [p*d.size//blox for p in range(blox)]
            bvec.append(d.size)
            for j in range(blox):
                part = d.toklist[bvec[j]:bvec[j+1]]
                if t in part:
                    tf += 1
                    for sw in range(1,slabsize+1):  ##  slab width
                        pt = part.index(t)  ##  t is present, so won't fail
                        p0 = pt - sw
                        if p0 >= 0:
                            pred = tuple(part[p0:pt])
                            pf = predtab.get(pred,0)
                            predtab[pred] = pf + 1
                        succ = tuple(part[pt+1:pt+sw+1])
                        if len(succ) == sw:
                            sf = succtab.get(succ,0)
                            succtab[succ] = sf + 1

    ##  evaluate predecessors & successors :
    newfreq = tf
    ##  targfreq = tf - math.log(tf)  ##  looser could be sqrt ??
    targfreq = tf ** 0.5
    ##  targfreq = max(math.log(tf),2)
    ##  targfreq = tf**0.5 + math.log(tf)
    ##  targfreq = max(2,targfreq*0.5)
    predlist = []
    for pred in predtab:
        n = predtab[pred]
        if n >= targfreq:
            r = round(n/tf,5)
            toks = len(pred)
            ##  a = alfs(pred)
            a = nats(paradat,pred)
            a2 = round(alfs(pred,1.0/Gold),5)
            ##  v = r * (toks+a2)  ##  v = r*(a+a2)
            v = r * a
            predlist.append((v,n,r,toks,a,a2,pred,t,tf))
    predlist.sort(reverse=True)
    succlist = []
    for succ in succtab:
        n = succtab[succ]
        if n >= targfreq:
            r = round(n/tf,5)
            toks = len(succ)
            ##  a = alfs(succ)
            a2 = round(alfs(succ,1.0/Gold),5)
            a = nats(paradat,succ)
            ##  v = r * (toks+a2)  ##  v = r*(a+a2)
            v = r * a
            succlist.append((v,n,r,toks,a,a2,t,succ,tf))
    succlist.sort(reverse=True)
    ##  for p in predlist:
        ##  print(p)
    ##  for s in succlist:
        ##  print(s)
    ##  decide on result :
    v1 = v2 = 0.0
    if predlist:
        bestpred = predlist[0]
        v1 = round(bestpred[0],7)
        toppred = bestpred[6]
    else:
        toppred = ()
    if succlist:
        bestsucc = succlist[0]
        v2 = round(bestsucc[0],7)
        topsucc = bestsucc[7]
    else:
        topsucc = ()

    ##  keep 1, both or just input node :
    toutlist = []
    if v1 >= cutvalue:
        if v2 >= cutvalue:
            ##  check whether to link both sides :
            keep = jointest(paradat,doclist,catname,toppred+(t,),(t,)+topsucc)
            ##  currently rather quick & dirty.
            if keep:  ##  link them
                vv = v1 * v2
                tout = toppred + (t,) + topsucc
                toutlist.append((round(vv,5),t,tout,v1,v2))
            else:  ##  dubious join
                if v1 > v2:
                    tout = toppred + (t,) ; vv = v1
                    toutlist.append((round(vv,5),t,tout,v1,v2))
                    if v2 >= cutvalue*Gold:  ##  keep both
                        tout = (t,) + topsucc
                        toutlist.append((round(v2,5),t,tout,v1,v2))
                else:
                    tout = (t,) + topsucc ; vv = v2
                    toutlist.append((round(vv,5),t,tout,v1,v2))
                    if v1 >= cutvalue*Gold:  ##  keep both
                        tout = toppred + (t,)
                        toutlist.append((round(v1,5),t,tout,v1,v2))                    
        else:
            ##  return pred
            tout = toppred + (t,)
            vv = v1
            toutlist.append((round(vv,5),t,tout,v1,v2)) 
    else:
        if v2 >= cutvalue:
            ##  return succ
            tout = (t,) + topsucc
            vv = v2
            toutlist.append((round(vv,5),t,tout,v1,v2)) 
        else:
            ##  return node only
            tout = (t,)
            vv = 0.0
            toutlist.append((round(vv,5),t,tout,v1,v2))
    
    return  toutlist


##  NATS :
def nats (paradat,toktuple):
    "sums info in nats of tokens in toktuple."

    nsum = 0.0
    dnom = paradat.tottoks + paradat.vocsize
    for t in toktuple:
        p = (paradat.allfrex.get(t,0)+1) / dnom
        nsum -= math.log(p)

    return  nsum


##  JOINTEST :
def jointest (paradat,doclist,catname,head,tail,minsize=144):
    "cheap & cheerful scoring of joinability."

    n = hits = 0
    hd = ' '.join(head) ; tl = ' '.join(tail) + ' '
    len1 = len(hd) ; len2 = len(tl)
    minsize = 60 ##  ; hz = tz = -99
    for j,d in enumerate(doclist):
        ##  dlen = len(text)
        ##  forward scan :
        ha = d.text.find(hd)
        ta = d.text.find(tl,ha+1,d.textlen)
        hit1 = ha >= 0 and ta >= 0 and abs(ta-ha) <= len1
        if ha > -1:
            n += 1
            if hit1:  ##  both found close 2gether
                hits += 1
        if d.textlen >= minsize:
            ## long enough; do backward scan :
            tz = d.text.rfind(tl)
            hz = d.text.rfind(hd,0,tz+len2+1)
            hit2 = hz >= 0 and tz >= 0 and abs(tz-hz) <= len1
            if ta == tz:
                i = 0.5  ##  same as leftward scan, discount
            else:
                i = 1
            if tz > -1:
                n += 1
                if hit2:  ##  both found close 2gether
                    hits += i
        ##  print(j,hits,n,text)
        ##  print(ha,ta,hz,tz)
    r = hits / n
    if r > 1.0-1.0/Gold:
        print(hd,"++++",tl,hits,n,r,file=paradat.dumpfile)
    else:
        print(hd,"----",tl,hits,n,r,file=paradat.dumpfile)

    return  r > 1.0-1.0/Gold


##  TESTJOIN :
def testjoin (doclist,catname,hd,tl):
    "heuristic to test whether head & tail should go together."

    s1 = ' '.join(hd) ; s2 = ' '.join(tl)
    n1 = len(s1) ; n2 = len(s2)
    hits = n = 0
    for d in doclist:
        if d.outcome == catname:
            p1 = d.text.find(s1)
            if p1 < 0:
                continue  ##  nowhere
            n += 1
            p2 = d.text.find(s2,p1+1,p1+n1+n2+2)
            if p2 > p1:
                hits += 1
            if d.size < 144:
                continue  ##  short text, only look once
            p2 = d.text.rfind(s2,p1,len(d.text)-1)
            if p2 < p1+n1+2:
                continue  ##  same example
            n += 1
            p1 = d.text.rfind(s1,p1+1,p2)
            if p1 > -1 and p1 > (p2-n1-n2-2):  ##  very close
                hits += 1

    r = hits / n

    if r >= 0.5:
        print(s1,"++++",s2,r)
    else:
        print(s1,"----",s2,r)

    return  r >= 0.5


##  JOINABLE :
def joinable (doclist,catname,hd,tl):
    "heuristic to test whether head & tail should go together."

    s1 = ' '.join(hd) ; s2 = ' '.join(tl)
    xvec = [] ; yvec = []
    for d in doclist:
        if d.outcome == catname:
            xvec.append(d.text.count(s1))
            yvec.append(d.text.count(s2))

    c = corrval(corsums(xvec,yvec))  ##  high if usually together

    if c > 0.7071:
        print(s1,"++++",s2,c)
    else:
        print(s1,"----",s2,c)

    return  c > 0.7071


##  OLDTAIL :
def oldtail():
    "previous ending of termext(); kept as backup."
    
    toppred = () ; pn = 0
    for pred in predtab:
        n = predtab[pred]
        if n >= targfreq:  ##  frequent enough
            if len(pred) > len(toppred):  ##  longer than current candidate
                toppred = pred ; pn = n
            elif len(pred) >= len(toppred) and (n > pn):
                toppred = pred ; pn = n
    if toppred:
        newfreq = pn
    topsucc = () ; sn = 0
    for succ in succtab:
        n = succtab[succ]
        if n >= targfreq:
            if len(succ) > len(topsucc):
                topsucc = succ ; sn = n
            elif len(succ) >= len(topsucc) and (n > sn):
                topsucc = succ ; sn = n
    if topsucc:
        newfreq = min(newfreq,sn)

    tout = tuple(toppred + (t,) + topsucc)

    return  (newfreq,tout,tf)    


##  ALFS :
def alfs (s,p=1):
    "number of alphabetics in a string or sequence."

    a = 0
    if type(s) == type('s'):
        for ch in s:
            if ch.isalpha():
                a += 1
    else:  ##  sequence
        for item in s:
            a += alfs(item)**p

    return  a




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
